/*
 * Copyright (c) 2005 Ichiro Satoh
 * All Rights Reserved.
 * written by Ichiro Satoh
 *
 * This program was developed by Ichiro Satoh at National Institute 
 * of Informatics as a part of a mobile object system, titled 
 * "AgentSpace".  All the programs of the system is developed by 
 * Ichiro Satoh and all rights are reserved by Ichiro Satoh. 
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the
 * copyright notice and this permission notice appear in all copies of
 * the software, derivative works or modified versions, and any
 * portions thereof, and that both notices appear in supporting 
 * documentation.
 */

/**
 * @author Ichiro Satoh
 * Created on 2005/05/4
 */

package agentspace;

import java.io.IOException;
import java.net.Inet6Address;
import java.net.InetAddress;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URI;
import java.net.UnknownHostException;

import agentspace.context.ContextManager;
import agentspace.discovery.AgentServerInfo;
import agentspace.discovery.HeartbeatManager;
import agentspace.discovery.RuntimeRegistry;
import agentspace.network.AgentSocketServer;
import agentspace.registry.AgentRegistry;
import agentspace.runtime.AgentRuntimeManager;
import agentspace.runtime.AgentRuntimeManagerImpl;
import agentspace.shell.AgentShell;

public class AgentServer {
	final static public int SERVER_PORT = 7000;
	final static public int HEARTBEAT_PORT = 7001;
	static InetAddress addr;
	static String host = null;
	static int port;
	static int heartbeatPort;
	static boolean autonomy = true;
	static boolean visual = true;
	static URL targetURL = null;
	static int maxSessions = 10;
	static int maxAliveSessions = 10;
	protected AgentServerInfo asi;
	protected HeartbeatManager discovery;

	static {
		URL.setURLStreamHandlerFactory(new agentspace.network.AgentURLStreamHandlerFactory());
	}

	public static void main(String[] args) {
		String command = null;
		port = SERVER_PORT;
		heartbeatPort = HEARTBEAT_PORT;
		for (int i = 0; i < args.length; i++) {
			if ((args[i].equals("-port") || args[i].equals("-p"))
					&& (i + 1 < args.length)) {
				try {
					port = Integer.valueOf(args[++i]).intValue();
				} 
				catch (NumberFormatException ex) {
					System.exit(1);
				}
			} 
			else if ((args[i].equals("-heartbeat") || args[i].equals("-b"))	&& (i + 1 < args.length)) {
				try {
					heartbeatPort = Integer.valueOf(args[++i]).intValue();
				} 
				catch (NumberFormatException ex) {
					ex.printStackTrace();
				}
			}
			else if (args[i].equals("-target") && (i + 1 < args.length)) {
				try {
					targetURL = new URL(args[++i]);
				} 
				catch (MalformedURLException ex) {
					ex.printStackTrace();
				}
			}
			else if (args[i].equals("-noautonomy")) {
				autonomy = false;
			}
			else if (args[i].equals("-cl")) {
				visual = false;
			}
			else if (args[i].equals("-MS") || args[i].equals("-MaxSessions")) {
				maxSessions  =Integer.valueOf(args[++i]).intValue();
			}
		}
		if (targetURL == null) {
			try {
				targetURL = new URL("matp://localhost:"+SERVER_PORT);
			} 
			catch (MalformedURLException e) {
				e.printStackTrace();
			}
		}
		AgentServer as = new AgentServer(command, port);
	}

	AgentServer(String command, int port) {
		System.out.println("Warning: java -Djava.net.preferIPv4Stack=true agentspace.AgentServer");
		String message = "AgentServer starting";
		try {
			addr = InetAddress.getLocalHost();
			host = addr.getHostAddress();
			message = message + " " + addr + ":"+ port;
			if (addr instanceof Inet6Address) {
				message = message + " (IPv6)";
			}
			else {
				message = message + " (IPv4)";
			}
		} 
		catch (UnknownHostException e) {
			e.printStackTrace();
			System.exit(0);
		}
		System.out.println(message);
		if (command == null || command.length() == 0) {
			try {
				RuntimeRegistry runtimes = new RuntimeRegistry();
				AgentRegistry registry = new AgentRegistry();
				AgentRuntimeManager manager = new AgentRuntimeManagerImpl(registry);
				ContextManager contextManager = new ContextManager(manager, registry);
				if (visual) {
					agentspace.monitor.AgentView view = new agentspace.monitor.AgentView(runtimes, contextManager);
					manager.addAgentRuntimeListener(view);
				}
				AgentShell shell = new AgentShell(contextManager, command);
				shell.start();
				AgentSocketServer server = new AgentSocketServer(manager, port);
				server.start();
				URI uri = new URI("matp://"+host+":"+port);
				if (visual) {
					agentspace.monitor.network.AgentNetworkView networkView = new agentspace.monitor.network.AgentNetworkView(uri); 
					runtimes.addRuntimeRegistryListener(networkView);
				}
				if (autonomy) {
					System.out.println("AgentServer: uri="+uri+" heartbeat_port="+heartbeatPort);
					discovery = new HeartbeatManager(runtimes, addr, port, heartbeatPort);
					discovery.start();
				}
			}
			catch (URISyntaxException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			catch (IOException ex) {
				ex.printStackTrace();                
			}
		}
	}

	static public InetAddress getServerInetAddress() throws UnknownHostException {
		return InetAddress.getLocalHost();
	}

	static public int getDefaultPort() {
		return port;
	}

	static public String getLocalHost() {
		return host;
	}

	static public URL getTargetURL() {
		return targetURL;
	}
}