package agentspace.discovery.scrap;

import java.io.IOException;
import java.net.InetAddress;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import agentspace.discovery.RuntimeDiscoveryExpire;
import agentspace.packet.Packet;
import agentspace.packet.SizeLimitExceededPacketException;
import agentspace.packet.multicast.MulticastUDPReceiver;
import agentspace.packet.multicast.MulticastUDPSender;
import agentspace.registry.AlreadyBoundException;
import agentspace.registry.NotBoundException;

public class MulticastPacketManager {
	final static protected String GROUPNAME = "228.1.2.3";
	protected URI uri;
	protected int port;
	protected boolean active = true;
	protected InetAddress localAddress = null;
	protected MulticastUDPSender sender;
	protected Thread receiverThread = null;
	protected MulticastUDPReceiver receiver;
	protected RuntimeDiscoveryExpire expire;
	protected InetAddress localhost;
	protected long serial = 0;
	protected Map<String,List<PacketListener>> table = null;

	public MulticastPacketManager(int port) throws IOException {
		this(port, GROUPNAME);
	}

	public MulticastPacketManager(int port, String groupName) throws IOException {
		this.port = port;
		localAddress = InetAddress.getLocalHost();
		receiver = new MulticastUDPReceiver(groupName, port);
		sender = new MulticastUDPSender(groupName, port);
	}

	public void multicast(Packet packet) throws IllegalArgumentException, IOException, SizeLimitExceededPacketException {
		packet.setSource(localAddress);
		packet.setReplySerial(serial++);
		sender.send(packet.getBytes());
	}

	Runnable executeReceiving = new Runnable() {
		public void run() {
			while (active) {
				try {
					byte[] data = receiver.receive();
					Packet packet = new Packet(data);
					String protocol = packet.getProtocol();
					String method = packet.getMethod();
					String key = protocol+":"+method;
					List<PacketListener> list = table.get(key);
					if (list != null) {
						synchronized(list) {
							for (PacketListener listener: list) {
								(new Thread(new Handler(listener, packet))).start(); 
							}
						}
					}
				}	
				catch (SizeLimitExceededPacketException e) {
					e.printStackTrace();
				}
				catch (URISyntaxException e) {
					e.printStackTrace();
				}
				catch (IOException e) {
					e.printStackTrace();
				}
			}
		}
	};

	class Handler implements Runnable {
		PacketListener listener;
		Packet packet;
		public Handler(PacketListener listener, Packet packet) {
			this.listener = listener;
			this.packet = packet;
		}
		public void run() {
			listener.received(packet);
		}
	}

	public synchronized void addPacketListener(PacketListener listener, String protocol, String method) throws AlreadyBoundException {
		String key = protocol + ":" + method;
		if (table == null) {
			table = new  LinkedHashMap<String, List<PacketListener>>();  
		}
		List<PacketListener> list = table.get(key);
		if (list == null) {
			list = new LinkedList<PacketListener>();
			list.add(listener);
			table.put(key, list);
		}
		else {
			synchronized(list) {
				if (list.contains(listener)) {
					throw new AlreadyBoundException(key+", "+listener);
				}
				list.add(listener);
			}
		}
	}

	public synchronized void removePacketListener(PacketListener listener, String protocol, String method) throws NotBoundException {
		String key = protocol + ":" + method;
		if (table == null) {
			throw new NotBoundException();
		}
		List<PacketListener> list = table.get(key);
		if (list == null) {
			throw new NotBoundException();
		}
		synchronized(list) {
			list.remove(listener);
			if (list.isEmpty()) {
				table.remove(key);
			}
		}
	}

	public void start() {
		active = true;
		if (receiverThread == null) {
			receiverThread = new Thread(executeReceiving);
			receiverThread.start();
		}
	}

	public void stop() {
		active = false;
		receiverThread.interrupt();
		receiverThread = null;
	}
}

