package agentspace.packet;

import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.StringTokenizer;

public class Packet {
	protected String protocol = null;
	protected String version = null;
	protected String method = null;
	protected URI uri = null;
	protected String content = null;
	public static final int PACKET_SIZE_LIMIT = 65515;
	protected HashMap<String, String> attributes = null;
	protected InetAddress source = null;
	protected InetAddress replyAddress = null;
	protected int replyPort = -1;
	protected long replySerial = -1;
	protected String serial = null;
	protected Exception ex = null;

	public Packet() {}

	public Packet(byte[] data) throws IOException, SizeLimitExceededPacketException, URISyntaxException {
		this.parse(data);
	}

	public String getMethod() {
		return method;
	}

	public void setMethod(String method) {
		this.method = method;
	}

	public URI getUri() {
		return uri;
	}

	public void setUri(URI uri) {
		this.uri = uri;
	}

	public String getProtocol() {
		return protocol;
	}

	public void setProtocol(String protocol) {
		this.protocol = protocol;
	}

	public String getVersion() {
		return version;
	}

	public void setVersion(String version) {
		this.version = version;
	}

	public String getContent() {
		return content;
	}

	public void setContent(String content) {
		this.content = content;
	}

	public void setAttribute(String key, String value) {
		if (attributes == null) {
			attributes = new HashMap<String, String>();
		}
		attributes.put(key.toUpperCase().trim(), value.trim());
	}

	public HashMap<String,String> getAttributes() {
		return attributes;
	}

	public String getAttribute(String name) {
		if (attributes == null) {
			return null;
		}
		return (String)attributes.get(name.toUpperCase());
	}
	
	public void setSource(InetAddress source) {
		this.source = source;
	}

	public InetAddress getSource() {
		return source;
	}

	public void setSerial(String serial) {
		this.serial = serial;
	}

	public String getSerial() {
		return serial;
	}

	public InetAddress getReplyAddress() {
		return replyAddress;
	}

	public void setReplyAddress(InetAddress replyAddress) {
		this.replyAddress = replyAddress;
	}

	public int getReplyPort() {
		return replyPort;
	}

	public void setReplyPort(int replyPort) {
		this.replyPort = replyPort;
	}

	public long getReplySerial() {
		return replySerial;
	}

	public void setReplySerial(long replySerial) {
		this.replySerial = replySerial;
	}

	public String toString() {
		String s = "";
		if (protocol != null) {
			s = s + "Protocol:"+protocol+", ";
		}
		if (method != null) {
			s = s + "Method:"+method+", ";
		}
		if (version != null) {
			s = s + "Version:"+version+", ";
		}
		if (uri != null) {
			s = s + "URI:"+uri+", ";
		}
		if (source != null) {
			s = s + "Source:"+source+", ";
		}
		if (serial != null) {
			s = s + "Serial:"+serial+", ";
		}
		if (replyAddress != null) {
			s = s + "Reply Address:"+replyAddress+", ";
		}
		if (replyPort > 0) {
			s = s + "Reply Port:"+replyPort+", ";
		}
		s = s + "[";
		if (attributes != null) {
			Iterator<String> i = attributes.keySet().iterator();
			while (i.hasNext()) {
				String key = i.next();
				s = s + "["+ key + ":" + attributes.get(key)+"]";
			}
		}
		s = s + "], ";
		if (content != null) {
			while (content.endsWith("\r") || content.endsWith("\n")) {
				content = content.substring(0, content.length() - 1);
			}
			s = s + "Content:"+content;
		}
		return s ;		
	}

	public String factory() {
		//		System.out.println("AgentUDPPacket: factory: "+this);

		if (method == null || uri == null ||  protocol == null || version == null) {
			throw new IllegalArgumentException("Illegal Argument: method:"+method+", uri:"+uri+", protocol:"+protocol+", version:"+version);
		}		    	
		String message = method + " " + uri + " " + protocol + "/" + version + "\r\n";
		if (source != null) {
			message = message + "SOURCE:" + source + "\r\n";
		}
		if (replyAddress != null && replyPort > 0) {
			message = message + "REPLY:" + replyAddress + ":" + replyPort + "\r\n";
		}
		if (attributes != null && attributes.size() > 0) {
			for (Iterator<Map.Entry<String,String>> it = attributes.entrySet().iterator() ; it.hasNext() ; ) {
				Map.Entry<String,String> entry = it.next();
				if (!entry.getKey().startsWith("SIZE")) {
					message = message + entry.getKey() + ":" + entry.getValue() + "\r\n";
				}
			}
		}
		if (content != null) {
			message = message + "SIZE:" + content.length() + "\r\n";
			message = message + "\r\n";
			message = message + content  + "\r\n";
		}
		message = message + "\r\n";
		return message;
	}

	public byte[] getBytes() throws IllegalArgumentException, SizeLimitExceededPacketException {
		byte[] data = null;
		try {
			ByteArrayOutputStream baos = new ByteArrayOutputStream(); 
			DataOutputStream dos = new DataOutputStream(baos);
			dos.writeBytes(factory());
			dos.close();
			data = baos.toByteArray();
			if (data.length > PACKET_SIZE_LIMIT) {
				throw new SizeLimitExceededPacketException("The size of packet: "+data.length);
			}
		}
		catch (IOException ex) {
			ex.printStackTrace();
		}
		return data;
	}

	public void parse(byte[] data) throws IOException, SizeLimitExceededPacketException, URISyntaxException {
		String message = new String(data);
		if (data.length > PACKET_SIZE_LIMIT) {
			throw new SizeLimitExceededPacketException("The size of packet: "+data.length);
		}
		int pos = message.indexOf('\n');  
		if (pos > 0) {
			String header = message.substring(0, pos);
			message = message.substring(pos+1, message.length());
			if (header.endsWith("\r")) {
				header = header.substring(0,  header.length() - 1);
			}
			StringTokenizer tokens = new StringTokenizer(header, " \t");
			method = tokens.nextToken();
			uri = new URI(tokens.nextToken());
			StringTokenizer st = new StringTokenizer(tokens.nextToken().trim(), "/");
			protocol = st.nextToken().trim().toUpperCase();
			version = st.nextToken().trim();
		}
		while (true) {
			int p = message.indexOf('\n');
			String attributeLine = message.substring(0,  p);
			message = message.substring(p+1, message.length());
			if (attributeLine.startsWith("\r\n\r\n")) {
				break;				
			}
			if (attributeLine.endsWith("\r")) {
				attributeLine = attributeLine.substring(0,  attributeLine.length() - 1);
			}
			if (attributeLine.length() == 0) {
				break;
			}
			int p_colon = attributeLine.indexOf(':');
			if (p_colon > 0) {
				setAttribute(attributeLine.substring(0, p_colon), attributeLine.substring(p_colon+1));
			}
		}
		if (message != null && message.length() > 0) {
			content = message;
			while (content.endsWith("\r\n")) {
				content = content.substring(0, content.length() - 2);
			}
		}
	}

	/**
	 * @param args
	 * @throws IOException 	
	 * @throws SizeLimitExceededPacketException 
	 * @throws IllegalArgumentException 
	 */
	public static void main(String[] args) throws IOException, IllegalArgumentException, SizeLimitExceededPacketException, URISyntaxException {
		// TODO Auto-generated method stub
		Packet message = new Packet();
		message.setMethod("ADVERTISE");
		//		message.setAddress(InetAddress.getLocalHost().getHostAddress());
		message.setUri(new URI("matp://localhost:7000"));
		message.setProtocol("AGENTSPACE");
		message.setVersion("0.3");
		message.setAttribute("STATUS", "NORMAL");
		message.setAttribute("INTERVAL", "2000");
		message.setContent("Hello\nWorld");
		System.out.println("Packet: message:\n=================\n"+message.factory()+"===============\n");
		byte[] data1 = message.getBytes();
		//		ByteArrayInputStream buf = new ByteArrayInputStream(data);

		Packet anotherMessage = new Packet(data1);
		System.out.println("Packet: anotherMessage: "+anotherMessage);
		System.out.println("Packet: parse("+anotherMessage+")");
		System.out.println("Packet: factory\n=================\n"+anotherMessage.factory()+"================");
		byte[] data2 = anotherMessage.getBytes();
		//		ByteArrayInputStream buf = new ByteArrayInputStream(data);

		Packet moreanotherMessage = new Packet(data2);
		System.out.println("Packet: more anotherMessage: "+moreanotherMessage);
		System.out.println("Packet: parse("+moreanotherMessage+")");
		System.out.println("Packet: factory\n=================\n"+moreanotherMessage.factory()+"================");

	}
}
