/*
 * Copyright (c) 2005, 2020 Ichiro Satoh
 * All Rights Reserved.
 * written by Ichiro Satoh
 *
 * This program was developed by Ichiro Satoh at National Institute 
 * of Informatics as a part of a mobile object system, titled 
 * "AgentSpace".  All the programs of the system is developed by 
 * Ichiro Satoh and all rights are reserved by Ichiro Satoh. 
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the
 * copyright notice and this permission notice appear in all copies of
 * the software, derivative works or modified versions, and any
 * portions thereof, and that both notices appear in supporting 
 * documentation.
 */

/**
 * @author Ichiro Satoh
 * Created on 2005/05/4, modified on 2020/7/19
 */

package agentspace.runtime;

import java.text.ParseException;
import java.util.Map;

import agentspace.AgentIdentifier;
import agentspace.Context;
import agentspace.runtime.loader.HashtableClassLoader;

public class AgentContainer {
	private int status;
	private String name;
	private AgentIdentifier id;
	private Class<?> cls;
	private Object obj;
	private long time;
	private HashtableClassLoader classLoader;
	private Map<String, byte[]> codes;
	private Map<String, byte[]> resources;
	private String agentClassName;
	private byte[] state;

	/* The following objects must to be transient because they must not be serialized */
	transient private AgentLifecycleManager manager;
	transient private ThreadGroup threadGroup;
	transient private Context context;
	
	public AgentContainer() {
		
	}
	
	public String getAgentClassName() {
		return agentClassName;
	}

	public void setAgentClassName(String agentClassName) {
		this.agentClassName = agentClassName;
	}

	public Class<?> getAgentClass() { 
		return cls;
	}
	
	public void setAgentClass(Class<?> cls) {
		this.cls = cls;
	}
	
	public Map<String, byte[]> getCodes() {
		return codes;
	}
	
	public void setCodes(Map<String, byte[]> codes) {
		this.codes = codes;
	}
	
	public Map<String, byte[]> getResources() {
		return resources;
	}
	
	public void setResources(Map<String, byte[]> resources) {
		this.resources = resources;
	}
	
	public void setTime(long time) {
		this.time = time;
	}
	
	public long getTime() {
		return time;
	}
	
	public String getTimeFormat() {
		java.text.DateFormat df = java.text.DateFormat.getTimeInstance(java.text.DateFormat.SHORT);
		return df.format(new java.util.Date(time));
	}
	
	public void setTimeFormat(String format) {
		try {
			java.text.DateFormat df = java.text.DateFormat.getTimeInstance(java.text.DateFormat.SHORT);
			time = df.parse(format).getTime();
		} 
		catch (ParseException e) {
			e.printStackTrace();
		}
	}
	
	public AgentIdentifier getIdentifier() {
		return id;
	}

	public void setIdentifier(AgentIdentifier id) {
		this.id = id;
	}

	public String getName() {
		if (name == null) {
			int pos = agentClassName.lastIndexOf('.');
			String n = agentClassName.substring(pos+1, agentClassName.length());
			return n;
		}
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}

	public int getStatus() {
		return status;
	}

	public void setStatus(int status) {
		this.status = status;
	}

	public Object getObject() {
		return obj;
	}

	public void setObject(Object obj) {
		this.obj = obj;
	}

	public HashtableClassLoader getClassLoader() {
		return classLoader;
	}

	public void setClassLoader(HashtableClassLoader classLoader) {
		this.classLoader = classLoader;
	}

	public AgentLifecycleManager getAgentLifetimeManager() {
		return manager;
	}

	public void setAgentLifetimeManager(AgentLifecycleManager manager) {
		this.manager = manager;
	}

	public byte[] getState() {
		return state;
	}

	public void setState(byte[] state) {
		this.state = state;
	}

	public ThreadGroup getThreadGroup() {
		return threadGroup;
	}

	public void setThreadGroup(ThreadGroup threadGroup) {
		this.threadGroup = threadGroup;
	}
	
	public Context getContext() {
		return context;
	}
	
	public void setContext(Context context) {
		this.context = context;
	}

	public boolean isAgent() {
		try {
			Class<?> agentClass = classLoader.loadClass(agentClassName);
			Class<?> typeClass = classLoader.loadClass("agentspace.Agent");
			return typeClass.isAssignableFrom(agentClass);
		} 
		catch (ClassNotFoundException e) {
			e.printStackTrace();
		}
		return false;
	}

	public boolean isTerminable() {
		if (isAgent()) {
			if ((status & AgentStatus.LIFE_MASK) ==  AgentStatus.NORMAL && 
					(status & AgentStatus.PROCESS_MASK) ==  AgentStatus.READY) {
				return true;
			}
		}
		return false;
	}
	
	public boolean isMobileAgent() {
		try {
			Class<?> agentClass = classLoader.loadClass(agentClassName);
			Class<?> typeClass = classLoader.loadClass("agentspace.Mobile");
			return typeClass.isAssignableFrom(agentClass);
		} 
		catch (ClassNotFoundException e) {
			e.printStackTrace();
		}
		return false;
	}

	public boolean isMobile() {
		if (isMobileAgent()) {
			if ((status & AgentStatus.LIFE_MASK) ==  AgentStatus.NORMAL && 
					(status & AgentStatus.PROCESS_MASK) ==  AgentStatus.READY) {
				return true;
			}
		}
		return false;
	}
	
	public boolean isPreservable() {
		if (isPreservableAgent()) {
			if ((status & AgentStatus.LIFE_MASK) ==  AgentStatus.NORMAL) {
				return true;
			}
		}
		return false;
	}

	public boolean isContainable() {
		if (isPreservableAgent()) {
			if ((status & AgentStatus.LIFE_MASK) ==  AgentStatus.NORMAL) {
				return true;
			}
		}
		return false;
	}

	public boolean isPreservableAgent() {
		try {
			Class<?> agentClass = classLoader.loadClass(agentClassName);
			Class<?> typeClass = classLoader.loadClass("agentspace.Preservable");
			return typeClass.isAssignableFrom(agentClass);
		} 
		catch (ClassNotFoundException e) {
			e.printStackTrace();
		}
		return false;
	}
	
	public boolean isDuplicatable() {
		if (isDuplicatableAgent()) {
			if ((status & AgentStatus.LIFE_MASK) ==  AgentStatus.NORMAL && 
					(status & AgentStatus.PROCESS_MASK) ==  AgentStatus.READY) {
				return true;
			}
		}
		return false;
	}
	
	public boolean isDuplicatableAgent() {
		try {
			Class<?> agentClass = classLoader.loadClass(agentClassName);
			Class<?> typeClass = classLoader.loadClass("agentspace.Duplicatable");
			return typeClass.isAssignableFrom(agentClass);
		} 
		catch (ClassNotFoundException e) {
			e.printStackTrace();
		}
		return false;
	}
	
	public String getType() {
		String type = "";
		if (isPreservableAgent()) {
			type = "preservable " + type;
		}
		if (isDuplicatableAgent()) {
			type = "duplicatable "+type;
		}
		if (isMobileAgent()) {
			type = "mobile " + type;
		}
		else {
			type = "stationary " + type;
		}
		return type; 	
	}
	
	public AgentContainer copyContainer() {
		AgentContainer ac = new AgentContainer();
		ac.setStatus(status);
		ac.setName(name);
		ac.setTime(time);
		ac.setCodes(codes);
		ac.setResources(codes);
		ac.setAgentClassName(agentClassName);
		return ac;
	}
	
	public String toString() {
		return name+" "+id+" "+AgentStatus.getLifeName(status)+" "+agentClassName;
	}
}
