/*
 * Copyright (c) 2005 Ichiro Satoh
 * All Rights Reserved.
 * written by Ichiro Satoh
 *
 * This program was developed by Ichiro Satoh at National Institute 
 * of Informatics as a part of a mobile object system, titled 
 * "AgentSpace".  All the programs of the system is developed by 
 * Ichiro Satoh and all rights are reserved by Ichiro Satoh. 
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the
 * copyright notice and this permission notice appear in all copies of
 * the software, derivative works or modified versions, and any
 * portions thereof, and that both notices appear in supporting 
 * documentation.
 */

/**
 * @author Ichiro Satoh
 * Created on 2005/05/4
 */

package agentspace;

import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import agentspace.dmi.Type;

public class Message implements Serializable, Cloneable {
	private static final long serialVersionUID = 6572677049411751858L;
	private String name  = null;
	private transient List<Arg> args  = null;
	private Future future = null;
	private int timeLimit = 0;
	private Object target = null;
	private Exception exception = null;

	/* for internal use in Message */
	class Arg {
		Byte type;
		Object value;
		public Arg(Byte type, Object value) {
			this.type = type;
			this.value = value;
		}			
		public Byte getType() {
			return type;
		}
		public Object getValue() {
			return value;
		}
		public String toString() {
			return "("+Type.toString(type)+":"+value+")";
		}
	}

	public Message() {
		args = new ArrayList<Arg>();
	}

	public Message(String name) {
		this();
		this.name = name;
	}

	public int getArgCount() {
		return args.size();
	}

	public String getName() {
		return name;
	}

	public void setArg(Object value) {
		args.add(new Arg(Type.TYPE_OBJECT, value));
	}

	public void setArg(boolean value) {
		args.add(new Arg(Type.TYPE_BOOLEAN, Boolean.valueOf(value))); 		
		
	}

	public void setArg(byte value) {
		args.add(new Arg(Type.TYPE_BYTE, Byte.valueOf(value))); 
	}

	public void setArg(short value) {
		args.add(new Arg(Type.TYPE_SHORT, Short.valueOf(value))); 
	}

	public void setArg(char value) {
		args.add(new Arg(Type.TYPE_CHAR, Character.valueOf(value))); 
	}

	public void setArg(int value) {
		args.add(new Arg(Type.TYPE_INT, Integer.valueOf(value))); 
	}

	public void setArg(long value) {
		args.add(new Arg(Type.TYPE_LONG, Long.valueOf(value))); 
	}

	public void setArg(float value) {
		args.add(new Arg(Type.TYPE_FLOAT, Float.valueOf(value))); 
	}

	public void setArg(double value) {
		args.add(new Arg(Type.TYPE_DOUBLE, Double.valueOf(value))); 
	}

	public Object[] getArgs() {
		Object[] objects = new Object[args.size()];
		for (int i = 0 ; i < args.size() ; i++ ) {
			objects[i] = args.get(i).getValue();
		}
		return objects;
	}

	public Class<?>[] getTypes() {
		Class<?>[] classes = new Class[args.size()];
		for (int i = 0; i < args.size(); i++) {
			switch (args.get(i).getType()) {
			case Type.TYPE_OBJECT:
				classes[i] = args.get(i).getValue().getClass();
				break;
			case Type.TYPE_BOOLEAN:
				classes[i] = java.lang.Boolean.TYPE;
				break;
			case Type.TYPE_BYTE:
				classes[i] = java.lang.Byte.TYPE;
				break;
			case Type.TYPE_SHORT:
				classes[i] = java.lang.Short.TYPE;
				break;
			case Type.TYPE_CHAR:
				classes[i] = java.lang.Character.TYPE;
				break;
			case Type.TYPE_INT:
				classes[i] = java.lang.Integer.TYPE;
				break;
			case Type.TYPE_LONG:
				classes[i] = java.lang.Long.TYPE;
				break;
			case Type.TYPE_FLOAT:
				classes[i] = java.lang.Float.TYPE;
				break;
			case Type.TYPE_DOUBLE:
				classes[i] = java.lang.Double.TYPE;
				break;
			case Type.TYPE_VOID:
				classes[i] = java.lang.Void.TYPE;
				break;
			default:
				throw new UnknownError("Invalid parameter type");	
			}
		}
		return classes;
	}

	public Object getTarget() {
		return target;
	}

	public void setFuture(Future futureObject) {
		future = futureObject;
	}

	public Future getFuture() {
		if (future == null) {
			future = new Future();
		}
		return future;
	}

	public void setTimelimit(int timelimit) {
		timeLimit = timelimit;
	}

	public int getTimelimit() {
		return timeLimit;
	}

	public void setException(Exception e) {
		if (exception == null) {
			exception = e;
		}
	}

	public Exception getException() {
		return exception;
	}

	private synchronized void writeObject(java.io.ObjectOutputStream out) throws IOException {
		out.writeObject(name);
		out.writeInt(args.size());
		for (int i = 0; i < args.size(); i++) {
			Object obj = args.get(i).getValue();
			switch (args.get(i).getType()) {
			case Type.TYPE_OBJECT:
				out.writeByte(Type.TYPE_OBJECT);
				out.writeObject(obj);
				break;
			case Type.TYPE_BOOLEAN:
				out.writeByte(Type.TYPE_BOOLEAN);
				out.writeBoolean(((Boolean)obj).booleanValue());
				break;
			case Type.TYPE_BYTE:
				out.writeByte(Type.TYPE_BYTE);
				out.writeByte(((Byte)obj).byteValue());
				break;
			case Type.TYPE_SHORT:
				out.writeByte(Type.TYPE_SHORT);
				out.writeShort(((Short)obj).shortValue());
				break;
			case Type.TYPE_CHAR:
				out.writeByte(Type.TYPE_CHAR);
				out.writeChar(((Character)obj).charValue());
				break;
			case Type.TYPE_INT:
				out.writeByte(Type.TYPE_INT);
				out.writeInt(((Integer)obj).intValue());
				break;
			case Type.TYPE_LONG:
				out.writeByte(Type.TYPE_LONG);
				out.writeLong(((Long)obj).longValue());
				break;
			case Type.TYPE_FLOAT:
				out.writeByte(Type.TYPE_FLOAT);
				out.writeFloat(((Float)obj).floatValue());
				break;
			case Type.TYPE_DOUBLE:
				out.writeByte(Type.TYPE_DOUBLE);
				out.writeDouble(((Double)obj).doubleValue());
				break;
			case Type.TYPE_VOID:
				out.writeByte(Type.TYPE_VOID);
				break;
			default:
				throw new UnknownError("Invalid parameter type");	
			}
		}
	}

	private synchronized void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException {
		name = (String)in.readObject();
		int size = in.readInt();
		args = new ArrayList<Arg>();
		for (int i = 0 ; i < size ; i++ ) {
			switch (in.readByte()) {
			case Type.TYPE_OBJECT:
				setArg(in.readObject());
				break;
			case Type.TYPE_BOOLEAN:
				setArg(in.readBoolean());
				break;
			case Type.TYPE_BYTE:
				setArg(in.readByte());
				break;
			case Type.TYPE_SHORT:
				setArg(in.readShort());
				break;
			case Type.TYPE_CHAR:
				setArg(in.readChar());
				break;
			case Type.TYPE_INT:
				setArg(in.readInt());
				break;
			case Type.TYPE_LONG:
				setArg(in.readLong());
				break;
			case Type.TYPE_FLOAT:
				setArg(in.readFloat());
				break;
			case Type.TYPE_DOUBLE:
				setArg(in.readDouble());
				break;
			case Type.TYPE_VOID:
				break;
			default:
				throw new UnknownError("Invalid parameter type");	
			}
		}
	}
	
	public String toString() {
		String s = "[" + getName() + ":";
		Iterator<Arg> i = args.iterator();
		while (i.hasNext()) {
			s = s + " " + i.next();
		}
		s = s + "]";
		return s;
	}
}
