package agentspace.discovery;

import java.io.IOException;
import java.net.InetAddress;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

import agentspace.packet.Packet;
import agentspace.packet.SizeLimitExceededPacketException;
import agentspace.packet.multicast.MulticastUDPReceiver;
import agentspace.packet.multicast.MulticastUDPSender;
import agentspace.registry.AlreadyBoundException;
import agentspace.registry.NotBoundException;

public class HeartbeatManager {
	//	final protected String groupName = "228.1.2.3";
	static final protected String groupName = "228.5.6.7";
	static final protected int EXPIRE = 3000;
	static final protected int CHECKING_INTERVAL = 200;
	static final protected int HEARTBEAT_INTERVAL = 1000;
	static final protected int HEARTBEAT_LIMIT = 3000;
//	static final protected int HEARTBEAT_INTERVAL = 2050;  for testing
//	static final protected int HEARTBEAT_LIMIT = 2000; for testing
	static final protected String INTERVAL = "INTERVAL";
	static final protected String HOST = "HOST";
	static final protected String MATP_HOST = "MATP-HOST";
	static final protected String MATP_PORT = "MATP-PORT";
	static final protected String HEARTBEAT_HOST = "HEARTBEAT-HOST";
	static final protected String HEARTBEAT_PORT = "HEARTBEAT-PORT";
	final int MAX_THREAD_NUMBER = 1;

	protected ScheduledExecutorService sendingExecutor = null;
	protected ExecutorService receivingExecutor = null;
	protected int matpPort = -1;   // Port for MATP
	protected int heartbeatPort = -1;   // Port for HEARTPORT
	protected InetAddress address;
	protected MulticastUDPSender sender = null;
	protected MulticastUDPReceiver receiver = null;
	protected RuntimeRegistry registry;
	protected ScheduledExecutorService monitorExecutor;
	protected List<ExpiredListener> heartbeatListeners;

	public static void main(String[] args) {
		int port = 7000;
		for (int i = 0; i < args.length; i++) {
			if ((args[i].equals("-port") || args[i].equals("-p"))
					&& (i + 1 < args.length)) {
				try {
					port = Integer.valueOf(args[++i]).intValue();
				} 
				catch (NumberFormatException ex) {
					System.exit(1);
				}
			}
		}
		RuntimeRegistry registry = new RuntimeRegistry();
		try {
			InetAddress addr = InetAddress.getLocalHost();
			HeartbeatManager manager = new HeartbeatManager(registry, addr, port, 7001);
			manager.start();
		}
		catch (IOException ex) {
			ex.printStackTrace();
		}
	}

	public HeartbeatManager(RuntimeRegistry regsitry, InetAddress address, int matpPort, int heartbeatPort) {
		this.registry = regsitry;
		this.address = address;
		this.matpPort = matpPort;
		this.heartbeatPort = heartbeatPort;
		receivingExecutor = Executors.newFixedThreadPool(MAX_THREAD_NUMBER);	
		sendingExecutor = Executors.newScheduledThreadPool(MAX_THREAD_NUMBER);
		/* for heartbeat monitor */
		monitorExecutor = Executors.newScheduledThreadPool(MAX_THREAD_NUMBER);
		heartbeatListeners = new ArrayList<ExpiredListener>();
	}

	public void shutdown() throws InterruptedException {
		receivingExecutor.shutdown();
		receivingExecutor.awaitTermination(30, TimeUnit.SECONDS);
		receivingExecutor.shutdownNow();
	}

	public void start()  throws IOException {
		if (receiver == null) {
			receiver = new MulticastUDPReceiver(groupName, heartbeatPort);
			receivingExecutor.execute(new PacketReceiver(registry, receiver));
		}
		//	    while (!receivingExecutor.isShutdown()) {
		//		}
		if (sender == null) {
			sender = new MulticastUDPSender(groupName, heartbeatPort);
			sendingExecutor.scheduleAtFixedRate(new HeartbeatSender(sender), HEARTBEAT_INTERVAL, HEARTBEAT_INTERVAL, TimeUnit.MILLISECONDS);
		}
		HeartbeatMonitor monitor = new HeartbeatMonitor(registry, heartbeatListeners);
		monitorExecutor.scheduleAtFixedRate(monitor, CHECKING_INTERVAL, CHECKING_INTERVAL, TimeUnit.MILLISECONDS);
	}

	class HeartbeatMonitor implements Runnable {
		boolean active = false;
		List<ExpiredListener> listeners;
		long lastTime = 0;

		public HeartbeatMonitor(RuntimeRegistry registry, List<ExpiredListener> listeners) {
			this.listeners = listeners;
			active = true;
		}

		public void run() {
			long currentTime = System.currentTimeMillis();
			Iterator<AgentServerInfo> itr = registry.iterator();
			while (itr.hasNext() && active) {
				AgentServerInfo asi = itr.next();
//				System.out.println("Monitor: checking for "+asi+", ("+(currentTime - asi.getLastTime()) +" <> "+HEARTBEAT_LIMIT+")");
				if (currentTime - asi.getLastTime() > HEARTBEAT_LIMIT) {
//					System.out.println("CheckingExpiredItem: already expired at "+asi+"("+(currentTime - asi.getLastTime()) +">"+HEARTBEAT_LIMIT+")");
					Iterator<ExpiredListener> j = listeners.iterator();
					while (j.hasNext()) {
						ExpiredListener listener = j.next();
						listener.expired(new ExpiredEvent(asi));
					}
					try {
						registry.remove(asi);
					}
					catch (NotBoundException e) {
						e.printStackTrace();
					}
				}
			}
		}
		
		public void stop() {
			active = false;
		}
	};

	class PacketReceiver implements Runnable {
		boolean active;
		MulticastUDPReceiver multicastReceiver;
		RuntimeRegistry registry;
		public PacketReceiver(RuntimeRegistry registry, MulticastUDPReceiver receiver) {
			this.multicastReceiver = receiver;
			this.registry = registry;
			active = true;
		}
		public void stop() {
			active = false;
		}
		public void run() {
			while(active) {
				try {
					byte[] data = multicastReceiver.receive();
//					System.out.println("Receiving at "+ (new Date()) + ":(" + data.length +")");
					Packet packet = new Packet(data);
					if (packet.getProtocol().equalsIgnoreCase("AGENTSPACE")) {
						URI uri = packet.getUri();
						if (uri != null) {
							if (packet.getMethod().equalsIgnoreCase("ADVERTISEMENT")) {
								String s = packet.getAttribute(HEARTBEAT_PORT);
								if (s != null) {
									int heartbeatPort = Integer.valueOf(s);
									String t = packet.getAttribute(INTERVAL);
									if (t != null) {
										long period = Long.valueOf(t);
										try {
											registry.add(uri, heartbeatPort, period);
//											System.out.println("Receiving: new uri="+uri+" heartbeatPort="+heartbeatPort+" interval="+period);
										}
										catch (AlreadyBoundException e) {
//											System.out.println("Receiving: already registered uri="+uri+" heartbeatPort="+heartbeatPort+" interval="+period);
										}
										AgentServerInfo asi = registry.lookup(uri);
										asi.setLastTime(System.currentTimeMillis());
									}
								}
							}
						}
						else if (packet.getMethod().equalsIgnoreCase("LEAVING")) {
							try {
								registry.remove(uri);
							} 
							catch (NotBoundException e) {
								e.printStackTrace();
							} 
						}
						else if (packet.getMethod().equalsIgnoreCase("LOOKUP")) {
						}
					}
				}
				catch (IOException ex) {
					ex.printStackTrace();
				} 
				catch (SizeLimitExceededPacketException ex) {
					ex.printStackTrace();
				} 
				catch (URISyntaxException ex) {
					ex.printStackTrace();
				} 
			}
		}
	};
	class HeartbeatSender implements Runnable {
		protected MulticastUDPSender sender;
		boolean active = true;
		String host;
		Packet packet;
		public HeartbeatSender(MulticastUDPSender sender) {
			this.sender = sender;
			try {
				host = String.valueOf(InetAddress.getLocalHost().getHostAddress());
			} 
			catch (UnknownHostException e) {
				e.printStackTrace();
			}
			packet = factory(host, matpPort, HEARTBEAT_INTERVAL);
//			System.out.println("SendingRunner: making packet = \n"+packet.factory());
		}
		public Packet factory(String host, int port, int interval) {
			Packet packet = new Packet();
			packet.setProtocol("AGENTSPACE");
			packet.setMethod("ADVERTISEMENT");
			packet.setVersion("0.1");
			packet.setAttribute(INTERVAL, String.valueOf(interval));
			try {
				packet.setUri(new URI("matp://"+host+":"+port));
			} 
			catch (URISyntaxException e) {
				e.printStackTrace();
			}
			packet.setAttribute(HEARTBEAT_PORT, String.valueOf(port));
			return packet;
		}

		public void stop() {
			active = false;
		}
		public void run() {
			try {
				sender.send(packet.getBytes());
//				System.out.println("SendingRunner: sent packet = "+packet);
			} 
			catch (IllegalArgumentException ex) {
				ex.printStackTrace();
			} 
			catch (IOException ex) {
				ex.printStackTrace();
			} 
			catch (SizeLimitExceededPacketException ex) {
				ex.printStackTrace();
			}
		}
	};
}