package agentspace.packet;

import java.io.IOException;
import java.io.Serializable;
import java.net.InetAddress;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.Date;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;

import agentspace.AgentIdentifier;
import agentspace.Message;
import agentspace.discovery.AgentServerInfo;
import agentspace.discovery.RuntimeDiscoveryExpire;
import agentspace.discovery.RuntimeRegistry;
import agentspace.discovery.scrap.Heartbeat;
import agentspace.packet.Packet;
import agentspace.packet.SizeLimitExceededPacketException;
import agentspace.packet.multicast.MulticastUDPReceiver;
import agentspace.packet.multicast.MulticastUDPSender;
import agentspace.registry.AgentRegistry;
import agentspace.registry.AlreadyBoundException;
import agentspace.registry.NotBoundException;

public class MulticastManager {
	final static protected String GROUPNAME = "228.1.2.3";
	final protected int EXPIRE = 3000;
	final protected int CHECKING_INTERVAL = 1000;
	final protected int SENDING_INTERVAL = 2000;
	//	final protected String PROTOCOL = "AGENTSPACE";
	//	final protected String VERSION = "0.3";
	//	final protected String ADVERTISE_METHOD = "ADVERTISE";
	//	final protected String LOOKUP_METHOD = "LOOKUP";
	protected URI uri;
	protected int port;
	protected boolean active = true;
	protected InetAddress localAddress = null;
	protected MulticastUDPSender sender;
	protected Thread receiverThread = null;
	protected MulticastManager manager;
	protected MulticastUDPReceiver receiver;
	protected RuntimeDiscoveryExpire expire;
	//	protected AgentRegistry registry;
	protected Packet packet;
	protected PacketExpireManager expireManager;
	protected long serial = 0;

	protected List heartbeats = null;
	protected Map<PacketListener,PacketListenerEntry> packetListeners = null;

	public MulticastManager(int port) throws IOException {
		this(port, GROUPNAME);		
	}

	public MulticastManager(int port, String groupName) throws IOException {
		this.port = port;
		localAddress = InetAddress.getLocalHost();
		receiver = new MulticastUDPReceiver(groupName, port);
		sender = new MulticastUDPSender(groupName, port);
	}

	public void multicast(Packet packet) throws IllegalArgumentException, IOException, SizeLimitExceededPacketException {
		packet.setSource(InetAddress.getLocalHost());
		packet.setReplySerial(serial++);
		sender.send(packet.getBytes());
	}

	public void startExpireManager(int interval) {
		if (expireManager == null) {
			throw new IllegalStateException();
		}			
		expireManager = new PacketExpireManager(interval);
		expireManager.start(); 
	}

	public Heartbeat startHeartbeat(Packet packet, int interval) {
		Heartbeat beat = new Heartbeat(new HeartbeatManager(sender, packet, interval), packet);
		if (heartbeats == null) {
			heartbeats = new LinkedList();		
		}
		heartbeats.add(beat);
		beat.start();
		return beat;
	}

	public void stopHeartbeat(Heartbeat beat) throws NotBoundException {
		if (beat == null) {
			throw new NotBoundException(beat.toString());
		}
		if (!heartbeats.contains(beat)) {
			throw new NotBoundException(beat.toString());
		}
		beat.stop();
		heartbeats.remove(beat);
	}

	class PacketListenerEntry {
		PacketListener packetLisener;
		String protocol;
		String method;
		long transaction;
		public PacketListenerEntry(PacketListener packetLisener, String protocol, String method, long transaction) {
			this.packetLisener = packetLisener;
			this.protocol = protocol;
			this.method = method;
			this.transaction = transaction;
		}
		public PacketListener getPacketLisener() {
			return packetLisener;
		}
		public String getProtocol() {
			return protocol;
		}
		public String getMethod() {
			return method;
		}
		public String get() {
			return method;
		}
	}

	public void addPacketListener(PacketListener packetListener, String protocol, String method, long serial) {
		PacketListenerEntry entry = new PacketListenerEntry(packetListener, protocol, method, serial);
		if (packetListeners == null) {
			packetListeners = new LinkedHashMap<PacketListener,PacketListenerEntry>(); 
		}
		packetListeners.put(packetListener, entry);		
	}

	public void removePacketListener(PacketListener packetListener) throws NotBoundException {
		PacketListenerEntry entry = (PacketListenerEntry)packetListeners.get(packetListener);
		if (entry == null) {
			throw new NotBoundException(); 
		}
		packetListeners.remove(packetListener);
	}

	Runnable executeReceiving = new Runnable() {
		public void run() {
			while (active) {
				try {
					byte[] data = receiver.receive();
					Packet packet;
					packet = new Packet(data);
					String protocol = packet.getProtocol();
					String method = packet.getMethod();
					long serial = packet.getReplySerial();
					if (packetListeners != null && packetListeners.size() > 0) {
						Iterator<PacketListenerEntry> i = packetListeners.values().iterator();
						while (i.hasNext()) {
							PacketListenerEntry entry = (PacketListenerEntry)i.next();
							if (protocol != null && protocol.equalsIgnoreCase(entry.getProtocol())) {
								if (method != null && method.equalsIgnoreCase(entry.getMethod())) {
									entry.getPacketLisener().received(packet);
								}
							}
						}
					}
				}	
				catch (SizeLimitExceededPacketException e) {
					e.printStackTrace();
				}
				catch (URISyntaxException e) {
					e.printStackTrace();
				}
				catch (IOException e) {
					e.printStackTrace();
				}
			}
		}
	};


	public PacketFuture lookup(Packet packet, int expire) throws IllegalArgumentException, IOException, SizeLimitExceededPacketException, URISyntaxException {
		PacketFuture future;
		packet.setReplyAddress(localAddress);
		packet.setReplyPort(port);
		packet.setReplySerial(serial);
		PacketListener listener = new PacketListener() {
			PacketFuture future;
			public PacketListener(PacketFuture future) {
				this.future = future;
			}
			public void received(Packet reply) {
				future.setReply(reply);
			}
		};
		addPacketListener(listener, packet.getProtocol(), "RESPONSE", serial);
		return future;
	}

	class PacketFuture {
		private boolean done = false;
		private boolean excepted = false;
		private Packet reply = null;
		private Exception ex = null;
		public PacketFuture(){}
		public synchronized void setReply(Packet reply) {
			done = true;
			this.reply = reply;
			notify();
		}
		public synchronized Packet getReply() throws Exception {
			while (done == false && excepted == false) {
				try{
					wait();
				}
				catch (InterruptedException e) {}
			}
			if (ex != null) {
				throw ex;
			}
			return reply;
		}

		public synchronized void setException(Exception ex) {
			excepted = true;
			this.ex = ex;
			notify();
		}
	}

	public void start() {
		active = true;
		//		expire.start();
		if (receiverThread == null) {
			receiverThread = new Thread(executeReceiving);
			receiverThread.start();
		}
		//		heartbeat.start();
	}

	public void stop() {
		active = false;
		//		heartbeat.stop();
		receiverThread.interrupt();
		receiverThread = null;
		//		expire.stop();
	}


}