/*
 * Copyright (c) 2005 Ichiro Satoh
 * All Rights Reserved.
 * written by Ichiro Satoh
 *
 * This program was developed by Ichiro Satoh at National Institute 
 * of Informatics as a part of a mobile object system, titled 
 * "AgentSpace".  All the programs of the system is developed by 
 * Ichiro Satoh and all rights are reserved by Ichiro Satoh. 
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the
 * copyright notice and this permission notice appear in all copies of
 * the software, derivative works or modified versions, and any
 * portions thereof, and that both notices appear in supporting 
 * documentation.
 */

/**
 * @author Ichiro Satoh
 * Created on 2005/05/5
 */

package agentspace.runtime;

import java.util.Stack;

public class AgentThreadPool {
	private Stack<Task> pool;
	private int size;
	private int counter;
	private String identify;
	private ThreadGroup group;

	/**
	 * Creates a new Pool instance
	 * 
	 * @param size
	 *            Max number of handler threads
	 * @param identify
	 *            Name of an Task instance
	 */
	public AgentThreadPool(ThreadGroup group, String identify, int size) {
		this.group = new ThreadGroup(group, "-POOL");
		this.identify = identify;
		this.size = size;
		pool = new Stack<Task>();
		counter = 1;
	}

	/**
	 * Requests the Pool to stop all the threads
	 * 
	 * @return void
	 */
	public void stopAllThreads() {
		group.interrupt();
	}

	/**
	 * Requests the Pool to perform some work.
	 * 
	 * @param handler
	 *            Handler to be dispatched.
	 * @return void
	 */
	public Thread startThread(Runnable handler) {
		Task task = null;
		synchronized (pool) {
			if (pool.empty()) {
				task = new Task(group, identify + (counter++), this);
				task.start();
			} else {
				task = (Task) pool.pop();
			}
		}
		task.wake(handler);
		return (Thread) task;
	}

	/**
	 * Convience method used by Task to put Thread back on the stack
	 * 
	 * @param task
	 *            Task to push
	 * @return boolean True if pushed, false otherwise
	 */
	protected boolean stopThread(Task task) {
		synchronized (pool) {
			if (pool.size() < size) {
				pool.push(task);
				return true;
			} else {
				return false;
			}
		}
	}

	class Task extends Thread {
		private Runnable task = null;

		private AgentThreadPool pool = null;

		public Task(String identify, AgentThreadPool pool) {
			super(identify);
			this.pool = pool;
		}

		public Task(ThreadGroup group, String identify, AgentThreadPool pool) {
			super(group, identify);
			this.pool = pool;
		}

		public synchronized void wake(Runnable task) {
			this.task = task;
			notify();
		}

		public synchronized void run() {
			boolean running = true;
			while (running) {
				try {
					if (task == null) {
						try {
							wait();
						} catch (InterruptedException e) {
							continue;
						}
					}
					if (task != null) {
						task.run();
					}
					task = null;
					running = pool.stopThread(this);
				} finally {
					task = null;
				}
			}
		}
	}

}